//
// Simple capture applications
//


/*
 * Header File Inclusion
 */
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include <string.h>
#include <linux/videodev2.h>
#include <linux/fb.h>
#include <linux/ti81xxfb.h>
#include <linux/ti81xxhdmi.h>
#include <linux/ti81xxvin.h>

/* device node to be used for capture */
#define CAPTURE_DEVICE      "/dev/video1"
#define CAPTURE_NAME        "Capture"
#define APP_NAME            "CaptureTest"

#define MAX_BUFFER              8
#define FRAME_TO_CAPTURE        10


unsigned long m_buffer_addr[MAX_BUFFER];

/* Structure for storing application data like file pointers, format etc. */
struct app_obj
{
    int fd;
    struct v4l2_capability cap;
    struct v4l2_format fmt;
    struct v4l2_dv_preset dv_preset;
    struct v4l2_requestbuffers reqbuf;
    int numbuffers;
    struct v4l2_buffer buf;
    struct ti81xxvin_overflow_status over_flow;
};

struct app_obj  m_cap;


/* Utility function for printing format */
static void printFormat(char *string, struct v4l2_format *fmt)
{
    printf("=============================================================\n");
    printf("%s Format:\n", string);
    printf("=============================================================\n");
    printf("fmt.type\t\t = %d\n", fmt->type);
    printf("fmt.width\t\t = %d\n", fmt->fmt.pix.width);
    printf("fmt.height\t\t = %d\n", fmt->fmt.pix.height);
    printf("fmt.pixelformat\t = %d\n", fmt->fmt.pix.pixelformat);
    printf("fmt.bytesperline\t = %d\n", fmt->fmt.pix.bytesperline);
    printf("fmt.sizeimage\t = %d\n", fmt->fmt.pix.sizeimage);
    printf("=============================================================\n");
}


void SetInput()
{
    int input_index = 0;

    int ret = ioctl(m_cap.fd, VIDIOC_S_INPUT, &input_index);

    if (ret)
    {
         printf("%s: Setting Input failed\n", APP_NAME);
         printf("Return Value = %d\n", ret);
         exit(2);
    }

}

//
// Query if present
//
void Query_DV_Present()
{

    m_cap.dv_preset.preset = -1;

    if (ioctl(m_cap.fd, VIDIOC_QUERY_DV_PRESET, &m_cap.dv_preset))
    {
         printf("Querying DV Preset failed\n");
         exit(2);
    }


    switch (m_cap.dv_preset.preset) {
    case V4L2_DV_720P60:
            printf("%s:\n Mode set is 720P60\n", APP_NAME);
            break;
    case V4L2_DV_1080I60:
            printf("%s:\n Mode set is 1080I60\n", APP_NAME);
            break;
    case V4L2_DV_1080P60:
            printf("%s:\n Mode set is 1080P60\n", APP_NAME);
            break;
    case V4L2_DV_1080P30:
            printf("%s:\n Mode set is 1080P30\n", APP_NAME);
            break;
    default:
           printf("%s:\n Failed to query preset\n", APP_NAME);
    }

}

void GetBufferFormat()
{

    /* Get the default format */
    m_cap.fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (ioctl(m_cap.fd, VIDIOC_G_FMT, &m_cap.fmt))
    {
          printf("Getting Buffer format failed");
          exit(2);
    }

    printFormat("Capture",&m_cap.fmt);


    /* Change required parameters and set the format */
    m_cap.fmt.fmt.pix.width = 640;
    m_cap.fmt.fmt.pix.height = 480;

    m_cap.fmt.fmt.pix.bytesperline = m_cap.fmt.fmt.pix.width * 2u;
    m_cap.fmt.fmt.pix.sizeimage = m_cap.fmt.fmt.pix.bytesperline * m_cap.fmt.fmt.pix.height;
    m_cap.fmt.fmt.pix.colorspace = V4L2_COLORSPACE_REC709;

    m_cap.fmt.fmt.pix.field = V4L2_PIX_FMT_YUYV;
    m_cap.fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;

    m_cap.fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    printf ("Calling set routine......\n");

    if (ioctl(m_cap.fd, VIDIOC_S_FMT, &m_cap.fmt))
    {
          printf("Setting format failed\n");
          exit(2);
    }

}


void RequestBuffers(int numbuffers)
{
    /* structure to store buffer request parameters */
    m_cap.reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    m_cap.reqbuf.count = numbuffers;
    m_cap.reqbuf.memory = V4L2_MEMORY_USERPTR;

    int ret = ioctl(m_cap.fd , VIDIOC_REQBUFS, &m_cap.reqbuf);
    if(ret < 0)
    {
        printf("cannot allocate memory\n");
        close(m_cap.fd);
        exit(4);
    }

}

//
//  start streaming
//
void StartStreaming()
{
    /* Start streaming */
    int cmd = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    int ret = ioctl(m_cap.fd, VIDIOC_STREAMON, &cmd);

    if (ret < 0)
    {
        perror("VIDIOC_STREAMON\n");
        exit(5);
    }
}

//
// Stop streaming
//
void StopStreaming()
{
    /* Stop streaming */
    int cmd = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    int ret = ioctl(m_cap.fd, VIDIOC_STREAMOFF, &cmd);

    if (ret < 0)
    {
        perror("VIDIOC_STREAMOFF\n");
        exit(6);
    }

}

void PrimeBuffers()
{
    /* Prime buffers */
    int ret;
    int i;

    int imageSize = m_cap.fmt.fmt.pix.sizeimage;

    printf("Image size = %d\n", imageSize);

    for (i = 0; i < MAX_BUFFER; i++)
    {
         m_cap.buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
         m_cap.buf.memory = V4L2_MEMORY_USERPTR;
         m_cap.buf.index = i;
         m_cap.buf.field = V4L2_FIELD_INTERLACED;


         int* buffer = (int*)calloc (1, imageSize);
         m_buffer_addr[m_cap.buf.index] = buffer;

         m_cap.buf.m.userptr = m_buffer_addr[m_cap.buf.index];
         m_cap.buf.length = imageSize;

         ret = ioctl( m_cap.fd, VIDIOC_QBUF, &m_cap.buf);

          if (ret < 0)
          {
              perror("VIDIOC_QBUF\n");
              return -1;
          }
    }

    printf("Prime buffer complete \n");


}


int CaptureData()
{
   int i;
   int ret;

   /* Start the loop of capture */
   for (i = 0; i < FRAME_TO_CAPTURE; i++)
   {
       /* Dequeue the captured buffer */
       ret = ioctl(m_cap.fd, VIDIOC_DQBUF, &m_cap.buf);
        if (ret < 0)
        {
            perror("VIDIOC_DQBUF\n");
             return -1;
        }

        // Process the buffer
        printf("Capturing data.....\n");


        /* Queue the buffer after processing is done */
        ret = ioctl(m_cap.fd, VIDIOC_QBUF, &m_cap.buf);

        if (ret < 0)
        {
              perror("VIDIOC_QBUF\n");
              return -1;
        }

   }
}

void free_buffers()
{
    /* Prime buffers */
    printf("Freeing buffers\n");
    int ret;
    int i;

    for (i = 0; i < MAX_BUFFER; i++)
    {
         int* buffer =   m_buffer_addr[i];
         free(buffer);
    }

}



int main()
{
    int mode = O_RDWR;

    /* Open capture driver */
    m_cap.fd = open((const char *)CAPTURE_DEVICE, mode);

    // Open the channel
    if (m_cap.fd != -1)
    {
        SetInput();
        Query_DV_Present();
        GetBufferFormat();

        RequestBuffers(MAX_BUFFER);
        PrimeBuffers();

        StartStreaming();
        CaptureData();
        StopStreaming();


        free_buffers();

        close(m_cap.fd);   // Close the channel
    }
    else
    {
        printf("failed to open capture device\n");
    }



}

